from __future__ import annotations

import os
from typing import TYPE_CHECKING

from pydantic import BaseModel
from pydantic.error_wrappers import ValidationError
from yaml import dump
from yaml import safe_load
from yaml import YAMLError

from cptk.utils import cptkException
if TYPE_CHECKING:
    from typing import TypeVar

    T = TypeVar('T')


class ConfigFileError(cptkException):
    """ Base cptkException for all errors thrown from the 'load_config_file'
    function, while trying to load a YAML configuration file. """


class ConfigFileNotFound(ConfigFileError, FileNotFoundError):
    def __init__(self, path: str) -> None:
        self.path = path
        super().__init__(f"Can't find configuration file {path!r}")


class ConfigFileValueError(ConfigFileError, ValueError):

    def __init__(self, path: str, errors: dict) -> None:
        self.path = path
        self.errors = errors
        super().__init__(self.__generate_error_message())

    def __generate_error_message(self) -> str:
        er = 'error' if len(self.errors) == 1 else f'{len(self.errors)} errors'
        s = f"{er} found in {self.path!r}:"

        for error in self.errors:
            path = '.'.join(error['loc']) if error['loc'] else '.'
            s += f"\nUnder {path!r}: {error['msg']}"

        return s


class ConfigFileParsingError(ConfigFileError):
    def __init__(
        self,
        path: str,
        error: str,
        position: tuple[int, int] = None,
    ) -> None:
        self.path = path
        self.error = error
        self.position = position
        super().__init__(self.__generate_error_message())

    def __generate_error_message(self) -> str:
        s = f"Error while parsing {self.path!r}\n"
        s += f"Under line {self.position[0]}: {self.error}"
        return s


class Configuration(BaseModel):

    @classmethod
    def load(cls: type[T], path: str) -> T:
        """ Load information from a YAML configuration file and dump it into a
        pydantic model. Raises relevent expections if the given file path isn't
        found, the YAML file can't be parsed, or the data doesn't match the
        pydantic model. """

        try:
            with open(path, encoding='utf8') as file:
                data = safe_load(file)

        except FileNotFoundError:
            raise ConfigFileNotFound(path)

        except YAMLError as err:
            mark = err.problem_mark
            pos = (mark.line + 1, mark.column + 1)
            raise ConfigFileParsingError(path, err.problem, pos)

        if not isinstance(data, dict):
            raise ConfigFileValueError(
                path, [{'loc': (), 'msg': "file isn't in dictionary format"}],
            )

        try:
            return cls(**data)
        except ValidationError as e:
            raise ConfigFileValueError(path, e.errors()) from e

    def yaml(self) -> str:
        """ Converts the object into a YAML string. """

        return dump(self.dict(exclude_unset=True), sort_keys=False)

    def dump(self, path: str) -> None:
        """ Dumps the pydantic model into the given file in YAML format. """

        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w', encoding='utf8') as file:
            file.write(self.yaml())



from .problem import Contest
from .problem import Problem
from .problem import Scraped
from .problem import Test
from .website import PageInfo
from .website import Website

__all__ = [
    'Contest',
    'Problem',
    'Scraped',
    'Test',
    'PageInfo',
    'Website',
]



from __future__ import annotations

import datetime
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING
from urllib import parse

from cptk.scrape import Contest
from cptk.scrape import Problem
from cptk.scrape import Test
from cptk.scrape import Website

if TYPE_CHECKING:
    from cptk.scrape import PageInfo
    from bs4 import BeautifulSoup


@dataclass(unsafe_hash=True)
class KattisProblem(Problem):
    # TODO: add author and source attributes.
    pass


@dataclass(unsafe_hash=True)
class KattisArchiveProblem(KattisProblem):
    difficulty: float = field(compare=True, default=None)


@dataclass(unsafe_hash=True)
class KattisContestProblem(KattisProblem):
    mark: str = field(compare=False, default=None)


@dataclass(unsafe_hash=True)
class KattisContest(Contest):
    pass


@dataclass(unsafe_hash=True)
class KattisArchive(Contest):
    pass


class Kattis(Website):

    @property
    def name(self) -> str:
        return 'Kattis'

    @property
    def domain(self) -> str:
        return 'open.kattis.com'

    def is_problem(self, info: PageInfo) -> bool:
        return info.data.find('div', {'class': 'problem-wrapper'}) is not None

    @staticmethod
    def _parse_code_text(soup: BeautifulSoup) -> None:
        text = soup.text.replace('<br>', '\n').replace('\r', '').strip()
        return '\n'.join(
            line.strip()
            for line in text.splitlines(keepends=False)
        ) + '\n'

    @classmethod
    def _parse_tests(cls, info: PageInfo) -> list[Test]:

        tables = info.data.find_all('table', {'class': 'sample'})
        res = list()

        for table in tables:
            inp, out = table.find_all('pre')
            res.append(
                Test(
                    input=cls._parse_code_text(inp),
                    expected=cls._parse_code_text(out),
                ),
            )

        return res

    def _parse_date(self, text: str) -> datetime.datetime:
        items = text.split()[:-1]  # last element is the timezone
        time = datetime.datetime.strptime(items.pop(), '%H:%M').time()
        date = (
            datetime.date.today() if not items
            else datetime.datetime.strptime(items.pop(), '%Y-%m-%d').date()
        )
        return datetime.datetime.combine(date, time)

    def _parse_contest(self, info: PageInfo) -> KattisContest | None:

        section = info.data.find('div', {'class': 'info upper'})
        if section is None:
            return None

        name = section.find('h2', {'class': 'title'}).text.strip()

        divs = section.find_all('div')
        start = divs[0].find('h4').next_sibling.text.strip()
        end = divs[-1].find('h4').next_sibling.text.strip()

        st = self._parse_date(start)
        et = self._parse_date(end)

        url = parse.urlparse(info.url)
        uid = url.path.split('/')[2]

        active = info.data.find(
            'div', {'class': 'contest-progress session-finished'},
        ) is None

        return KattisContest(
            _uid=uid,
            url=parse.urljoin(info.url, f'/contests/{uid}'),
            website=self,
            name=name,
            start_time=st,
            end_time=et,
            active=active,
        )

    def _parse_archive(self, info: PageInfo) -> KattisArchive:
        url = parse.urlparse(info.url)
        uid = url.path.split('/')[1]
        return KattisArchive(
            uid,
            website=self,
            name=uid.title(),
            url=parse.urljoin(info.url, f'/{uid}'),
        )

    def _parse_sidebar_field(self, info: PageInfo, field: str) -> str:
        sidebar = info.data.find(
            'div',
            {'class': 'problem-sidebar sidebar-info'},
        )

        metadatas = sidebar.find_all('p')
        field_soup = next(
            soup for soup in metadatas
            if field in soup.text
        )

        return field_soup.text.split(':')[-1].strip()

    def to_problem(self, info: PageInfo) -> KattisProblem:

        title = info.data.find('div', {'class': 'headline-wrapper'}).find('h1')
        br = title.find('br')
        name = br.next_sibling.text if br else title.text

        uid = self._parse_sidebar_field(info, 'Problem ID')
        time_limit = self._parse_sidebar_field(info, 'CPU Time limit')
        memory_limit = self._parse_sidebar_field(info, 'Memory limit')

        kwargs = {
            '_uid': uid.strip(),
            'website': self,
            'name': name.strip(),
            'url': info.url,
            'tests': self._parse_tests(info),
            'time_limit': int(time_limit.split()[0]),
            'memory_limit': int(memory_limit.split()[0]),
        }

        contest = self._parse_contest(info)

        if contest:
            problem_title = next(title.children)
            mark = problem_title.text.split()[-1]
            return KattisContestProblem(
                **kwargs,
                contest=contest,
                mark=mark.strip(),
            )

        else:
            archive = self._parse_archive(info)
            diff = float(self._parse_sidebar_field(info, 'Difficulty'))
            return KattisArchiveProblem(
                **kwargs,
                contest=archive,
                difficulty=diff,
            )



from __future__ import annotations

import os
import subprocess
import sys
from dataclasses import dataclass
from threading import Timer
from typing import TypeVar

import cptk.utils
from cptk.core.system import System
from cptk.local.problem import LocalProblem

T = TypeVar('T')


@dataclass
class RunnerResult:
    runner: Runner
    code: int
    timed_out: bool
    outs: str | None = None
    errs: str | None = None


class Nothing:
    def __getattribute__(self: T, *_) -> T:
        return self

    def __call__(self: T, *_, **__) -> T:
        return self


class BakingError(cptk.utils.cptkException):
    def __init__(self, code: int, cmd: str) -> None:
        self.code = code
        self.cmd = cmd
        super().__init__(
            f'Execution of command resulted in exit code {code}:\n{cmd}',
        )


class Runner:
    def __init__(self, env: dict = None) -> None:
        self.env = env if env is not None else os.environ

    def exec(
        self,
        cmd: str,
        input: str = None,
        timeout: float = None,
        redirect: bool = True,
        wd: str = None,
    ) -> RunnerResult:
        """ Executes the given command, and returns a 'RunnerResult' instance
        that describes the result of the execution.
        If input is provided, it is piped into the input of the subprocess.
        If timeout is provided, the execution of the process will get
        terminated after the provided amount of seconds. """

        proc = subprocess.Popen(
            cmd.split(),
            cwd=wd,
            env=self.env,
            encoding='utf8',
            stdin=subprocess.PIPE if redirect else sys.stdin,
            stdout=subprocess.PIPE if redirect else sys.stdout,
            stderr=subprocess.PIPE if redirect else sys.stderr,
        )

        # I used a timer thread and a Popen instance instead of the builtin
        # run or communicate methods with the 'timeout' argument to allow
        # capturing the stdin and stderr even if we terminate the process.

        timer = Timer(timeout, proc.kill) if timeout is not None else Nothing()

        if input is not None:
            proc.stdin.write(input)
            proc.stdin.flush()

        try:
            timer.start()
            proc.wait()
        finally:
            timed_out = False if timeout is None else not timer.is_alive()
            timer.cancel()

        return RunnerResult(
            runner=self,
            outs=proc.stdout.read() if redirect else None,
            errs=proc.stderr.read() if redirect else None,
            code=proc.returncode,
            timed_out=timed_out,
        )


class Chef:
    """ Bake, serve and test local problems. """

    def __init__(self, problem: LocalProblem) -> None:
        self._problem = problem
        self._runner = Runner()

    def bake(self) -> None:
        """ Bakes (generates) the executable of the current problem solution.
        If the recipe configuration file of the current problem doesn't specify
        a 'bake' option, returns None quietly. """

        location = self._problem.location

        for cmd in self._problem.recipe.bake:
            System.log(cmd)
            res = self._runner.exec(cmd, wd=location, redirect=False)
            if res.code:
                raise BakingError(res.code, cmd)

    def serve(self) -> None:
        """ Bakes the local problem (if a baking recipe is provided), and serves
        it while piping the standard input to the executable. """

        self.bake()

        cmd = self._problem.recipe.serve
        location = self._problem.location

        System.log(cmd)
        res = self._runner.exec(cmd, wd=location, redirect=False)
        System.abort(res.code)

    def test(self) -> None:
        """ Bakes (if a baking recipe is provided) and serves the local tests
        that are linked to the problem. """



from __future__ import annotations

import os
import platform
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING

import jinja2
from slugify import slugify

if TYPE_CHECKING:
    from cptk.scrape import Problem
    from typing import TypeVar
    T = TypeVar('T')

from cptk.utils import cptkException


class PreprocessError(cptkException, ABC):

    def __init__(self, error: Exception) -> None:
        self.jinja_error = error
        super().__init__(self._generate_error_message())

    @abstractmethod
    def _generate_error_message(self) -> str:
        pass


class PreprocessStringError(PreprocessError):
    def _generate_error_message(self) -> str:
        return ', '.join(self.jinja_error.args)


class PreprocessFileError(PreprocessError):
    def __init__(self, error: Exception, path: str) -> None:
        self.path = path
        super().__init__(error)

    def _generate_error_message(self) -> str:
        return f'In {self.path}: ' + ', '.join(self.jinja_error.args)

    @classmethod
    def from_string_err(
        cls: type[T],
        path: str,
        err: PreprocessStringError,
    ) -> T:
        return cls(err.jinja_error, path)


class PreprocessNameError(PreprocessError):
    def __init__(self, error: Exception, name: str) -> None:
        self.name = name
        super().__init__(error)

    def _generate_error_message(self) -> str:
        return f'In {self.name}: ' + ', '.join(self.jinja_error.args)

    @classmethod
    def from_string_err(
        cls: type[T],
        name: str,
        err: PreprocessStringError,
    ) -> T:
        return cls(err.jinja_error, name)


class Preprocessor:

    def __init__(self, problem: Problem) -> None:
        self._env = jinja2.Environment(undefined=jinja2.StrictUndefined)
        self._env.globals.update({
            key: val
            for key, val in {
                'problem': problem,
                'now': datetime.now(),
                'slug': slugify,
                'system': platform.system(),
                'user': self.__try(os.getlogin),
            }.items() if val is not None
        })

        # Slug is a global to allow the "slug(...)" syntax,
        # and a filter to allow the "... | slug" syntax.
        self._env.filters.update({'slug': slugify})

        # Cptk considers None values as undefined.
        # This means that scopes like {% if v is defined %} where v is None
        # won't be executed.

        def defined(v):
            return not isinstance(v, jinja2.Undefined) and v is not None

        def undefined(v):
            return isinstance(v, jinja2.Undefined) or v is None

        self._env.tests.update({'defined': defined, 'undefined': undefined})

    @staticmethod
    def __try(fun, default=None):
        try:
            return fun()
        except Exception:
            return default

    def parse_string(self, string: str) -> str:
        try:
            return self._env.from_string(string).render()
        except Exception as err:
            raise PreprocessStringError(err)

    def parse_file_contents(self, path: str) -> None:
        with open(path, encoding='utf8') as file:
            data = file.read()

        try:
            new = self.parse_string(data)
        except PreprocessStringError as err:
            raise PreprocessFileError.from_string_err(path, err)

        with open(path, 'w', encoding='utf8') as file:
            file.write(new)

    def parse_directory(self, path: str) -> None:
        for item in os.listdir(path):
            old = os.path.join(path, item)

            try:
                new = os.path.join(path, self.parse_string(item))
            except PreprocessStringError as err:
                raise PreprocessNameError.from_string_err(old, err)

            os.rename(src=old, dst=new)

            if os.path.isdir(new):
                self.parse_directory(new)

            elif os.path.isfile(new):
                self.parse_file_contents(new)



from __future__ import annotations

import sys
from typing import TYPE_CHECKING

import colorama

from cptk.utils import cptkException

if TYPE_CHECKING:
    from subprocess import CompletedProcess


class SystemRunError(cptkException):
    """ Raised by System.run if something goes wrong. """


class System:

    CMD = colorama.Fore.YELLOW
    LOG = colorama.Fore.LIGHTBLACK_EX
    ERROR = colorama.Back.RED + colorama.Style.BRIGHT
    WARN = colorama.Back.YELLOW + colorama.Fore.BLACK + colorama.Style.BRIGHT
    RESET = colorama.Style.RESET_ALL

    _verbosity = 0
    _yes_stack = 0

    @classmethod
    def run(
        cls,
        cmd: str,
        errormsg: str = None,
        verbose: bool = None,
    ) -> CompletedProcess:
        """ Runs the given command in the terminal. If 'errormsg' is provided,
        asserts that the returncode from the process is zero, and if not,
        raises an SystemRunError with the given message. If 'verbose' is
        provided, it overwrites the classes verbosity setting. """

        import subprocess

        if verbose is None:
            verbose = cls._verbosity >= 2

        if verbose:
            cls.echo(cls.CMD + cmd + cls.RESET)

        res = subprocess.run(
            cmd.split(),
            check=False,
            stdout=sys.stdout if verbose else subprocess.PIPE,
            stderr=sys.stderr if verbose else subprocess.PIPE,
        )

        if errormsg is not None and res.returncode != 0:
            raise SystemRunError(errormsg)

        return res

    @classmethod
    def set_verbosity(cls, level: int = 0) -> None:
        """ Verbosity can be True, False, or None (which means "defalt"). """
        cls._verbosity = level

    @classmethod
    def set_yes(cls, amount: int = 1) -> None:
        cls._yes_stack = amount

    @classmethod
    def pop_yes(cls) -> bool:
        if cls._yes_stack > 0:
            cls._yes_stack -= 1
            return True
        return False

    @staticmethod
    def _expection_to_msg(error: Exception) -> str:
        return ', '.join(str(a) for a in error.args)

    @classmethod
    def error(cls, error: str | Exception) -> None:
        if isinstance(error, Exception):
            error = cls._expection_to_msg(error)

        cls.echo(f"{cls.ERROR} ERROR {cls.RESET} {error}")

    @classmethod
    def unexpected_error(cls, error: Exception) -> None:

        title = type(error).__name__
        desc = cls._expection_to_msg(error)
        msg = title if not desc else f'{title}: {desc}'

        cls.echo(f"{cls.ERROR} UNEXPECTED ERROR {cls.RESET} {msg}")

        tb = error.__traceback__
        while tb is not None:
            file = tb.tb_frame.f_code.co_filename
            lineno = tb.tb_lineno
            cls.echo(f'Inside {cls.CMD}{file}:{lineno}{cls.RESET}')
            tb = tb.tb_next

    @classmethod
    def abnormal_exit(cls) -> None:
        cls.echo(f"{cls.ERROR} ABNORMAL EXIT {cls.RESET}")

    @classmethod
    def warn(cls, msg: str) -> None:
        cls.echo(f"{cls.WARN} WARNING {cls.RESET} {msg}")

    @classmethod
    def confirm(cls, question: str) -> bool:
        if cls.pop_yes():
            return True
        return cls.ask(question, {('Y', 'y'): True, ('n', 'N'): False})

    @classmethod
    def ask(cls, question: str, options: dict) -> bool:

        titles = list()
        answers = dict()
        for option, value in options.items():
            if isinstance(option, str):
                titles.append(option)
                answers[option] = value
            else:
                titles.append(option[0])
                for answer in option:
                    answers[answer] = value

        question = f'{question}? [{"/".join(titles)}]: '
        query = f"{cls.CMD}{question}{cls.RESET}"
        res = input(query).strip()

        while res not in answers:
            res = input(query).strip()
        return answers[res]

    @classmethod
    def abort(cls, code: int = 1) -> None:
        raise SystemExit(code)

    @classmethod
    def echo(cls, msg: str) -> None:
        print(msg)  # noqa: T001

    @classmethod
    def log(cls, msg: str) -> None:
        if cls._verbosity >= 1:
            cls.echo(cls.LOG + msg + cls.RESET)



from __future__ import annotations
RECIPE_FILE = 'recipes.cptk.yaml'
PROJECT_FILE = '.cptk/project.cptk.yaml'
LAST_FILE = '.cptk/stayaway/last.cptk.txt'

MOVE_FILE = '.cptk/moves.cptk.txt'
MOVE_FILE_SEPERATOR = '::'
MOVE_SAFES = ['./', '**/.cptk/**']

LAST_FILE_SEPERATOR = '::'



from __future__ import annotations

import os
import shutil
from dataclasses import dataclass
from dataclasses import field
from glob import iglob
from typing import TYPE_CHECKING

from pydantic import BaseModel

import cptk.constants
import cptk.utils
from cptk.core.config import ConfigFileParsingError
from cptk.core.config import Configuration
from cptk.core.fetcher import Fetcher
from cptk.core.preprocessor import Preprocessor
from cptk.core.system import System
from cptk.core.templates import DEFAULT_TEMPLATES
from cptk.local.problem import LocalProblem
from cptk.local.problem import Recipe


if TYPE_CHECKING:
    from typing import TypeVar
    from cptk.scrape import Problem
    from cptk.core.templates import Template
    T = TypeVar('T')


class ProjectNotFound(cptk.utils.cptkException):
    def __init__(self) -> None:
        super().__init__("Couldn't find a cptk project recursively")


class InvalidMovePath(cptk.utils.cptkException):
    def __init__(self, path: str, msg: str) -> None:
        self.path = path
        super().__init__(msg)


class InvalidMoveSource(InvalidMovePath):
    def __init__(self, path: str) -> None:
        super().__init__(path, f"Can't move {path!r}")


class InvalidMoveDest(InvalidMovePath):
    def __init__(self, path: str) -> None:
        super().__init__(path, f"Can't move to {path!r}")


class InvalidTemplate(cptk.utils.cptkException):
    pass


class CloneSettings(BaseModel):
    template: str
    path: str
    recipe: Recipe

    def dict(self, **kwargs) -> dict:
        kwargs.update({"exclude_unset": False})
        return super().dict(**kwargs)


class ProjectConfig(Configuration):
    clone: CloneSettings


@dataclass(unsafe_hash=True)
class LocalProject:
    location: str = field(compare=True)

    def __init__(self, location: str) -> None:
        self.location = location

    @cptk.utils.cached_property
    def fetcher(self) -> Fetcher:
        return Fetcher()

    @classmethod
    def is_project(cls, location: str) -> bool:
        """ Returns True if the given location is the root of a valid cptk
        project. """
        path = os.path.join(location, cptk.constants.PROJECT_FILE)
        return os.path.isfile(os.path.join(path))

    @classmethod
    def find(cls: type[T], location: str) -> T:
        """ Recursively searches if the given location is part of a cptk
        project, and if so, returns an instance of the project. If a project
        isn't found, an error is thrown. """

        if cls.is_project(location):
            return cls(location)

        parent = os.path.dirname(location)
        if parent == location:
            raise ProjectNotFound()

        return cls.find(parent)

    @classmethod
    def init(cls: type[T], location: str, template: str) -> T:
        """ Initialize an empty local project in the given location using the
        given template name. Returns the newly created project as a LocalProject
        instance. """

        avaliable_templates = {t.uid: t for t in DEFAULT_TEMPLATES}
        if template not in avaliable_templates:
            raise InvalidTemplate(f'Invalid template name {template!r}')

        template: Template = avaliable_templates.get(template)
        commons = cptk.utils.find_common_files(location, template.path)

        if commons:
            System.warn(
                '\n'.join((
                    'The following files will be overwritten:',
                    *commons,
                )),
            )

            ans = System.confirm('Are you sure you want to continue')
            if not ans:
                System.abort()

        cptk.utils.soft_tree_copy(src=template.path, dst=location)
        return cls(location)

    @cptk.utils.cached_property
    def config(self) -> ProjectConfig:
        p = os.path.join(self.location, cptk.constants.PROJECT_FILE)
        return ProjectConfig.load(p)

    def _load_moves(self) -> list[tuple[str, str]]:
        """ Loads information from the local moves file and returns the
        replacements directory. """

        moves_path = self.relative(cptk.constants.MOVE_FILE)

        try:
            with open(moves_path) as file:
                lines = file.read().splitlines(keepends=False)
        except FileNotFoundError:
            return list()

        moves = list()
        for lineno, line in enumerate(lines, start=1):
            parts = line.split(cptk.constants.MOVE_FILE_SEPERATOR)

            if len(parts) == 2:
                moves.append((parts[0], parts[1]))

            else:
                raise ConfigFileParsingError(
                    path=moves_path,
                    error=f'Seperator {cptk.constants.MOVE_FILE_SEPERATOR!r}'
                    ' not found',
                    position=(lineno, 0),
                )

        return moves

    @staticmethod
    def _is_subpath(parent: str, sub: str) -> bool:
        parent = os.path.abspath(parent)
        sub = os.path.abspath(sub)
        return os.path.commonpath([parent, sub]) == parent

    def _single_move(self, path: str, src: str, dst: str) -> str:
        """ If the given path shares a prefix with the move source, updates
        the path with by replacing the prefix with the destination prefix
        accordingly. If a prefix isn't shared, returns the original path
        without modifications. """

        if not self._is_subpath(src, path):
            return path
        rel = os.path.relpath(path, src)
        return os.path.normpath(os.path.join(dst, rel))

    def move_relative(self, path: str) -> str:
        """ Applies all registered move transformations to the given path, and
        returns the new path, as an abs path. """

        path = self.relative(path)
        for src, dst in self._load_moves():
            path = self._single_move(
                path,
                self.relative(src),
                self.relative(dst),
            )
        return path

    def relative(self, path: str) -> str:
        """ If the given path is not absolute, returns the absolute path relative
        to the project location. """

        if path is not None and not os.path.isabs(path):
            path = os.path.join(self.location, path)
        return path

    def clone_url(self, url: str) -> LocalProblem:
        """ Clones the given URL as a problem and stores as a local problem
        inside the current cptk project. """

        page = self.fetcher.to_page(url)
        problem = self.fetcher.page_to_problem(page)
        return self.clone_problem(problem)

    def clone_problem(self, problem: Problem) -> LocalProblem:
        """ Clones the given problem instance and stores a local problem inside
        the current cptk project. """

        processor = Preprocessor(problem)

        src = self.relative(self.config.clone.template)
        dst = self.move_relative(processor.parse_string(self.config.clone.path))
        commons = cptk.utils.find_common_files(src, dst)

        if commons:
            System.warn(
                '\n'.join((
                    'The following files will be overwritten:',
                    *commons,
                )),
            )

            ans = System.confirm('Are you sure you want to continue')
            if not ans:
                System.abort()

        cptk.utils.soft_tree_copy(src, dst)
        recipe = self.config.clone.recipe.preprocess(processor)
        prob = LocalProblem.init(dst, recipe)

        processor.parse_directory(dst)
        self.update_last(prob)
        return prob

    def last(self) -> LocalProblem | None:
        try:
            with open(self.relative(cptk.constants.LAST_FILE)) as file:
                data = file.read()
        except FileNotFoundError:
            return None

        location, _, name = data.partition(cptk.constants.LAST_FILE_SEPERATOR)
        return LocalProblem(location, name if name else None)

    def update_last(self, prob: LocalProblem) -> None:
        path = self.relative(cptk.constants.LAST_FILE)
        os.makedirs(os.path.dirname(path), exist_ok=True)
        with open(path, 'w') as file:
            location = prob.location
            name = '' if prob.name is None else prob.name
            file.write(f'{location}{cptk.constants.LAST_FILE_SEPERATOR}{name}')

    def move(self, src: str, dst: str) -> None:
        """ Validates that the given source and destination directories are
        can be moved, and if so, moves source to destination and record the
        move. After the move, new problem that share a prefix with the source
        will be created in the given destination with the matching suffix. """

        # normalize paths to their unique absolute representation
        src = os.path.normpath(self.relative(src))
        dst = os.path.normpath(self.relative(dst))

        if not self._is_subpath(self.location, src) or not os.path.isdir(src):
            raise InvalidMoveSource(src)

        if not self._is_subpath(self.location, dst):
            raise InvalidMoveDest(dst)

        for pat in cptk.constants.MOVE_SAFES:
            # This is an ugly solution. I have tries to use fnmatch but
            # it has some strange behavior and doesn't match directories
            # if their name doesn't end with an '/'. I should rewrite this
            # piece of code, but as long as it works and preforms ok-ish,
            # I'm fine with it. (:

            matches = (
                os.path.normpath(g)
                for g in iglob(self.relative(pat), recursive=True)
            )
            if os.path.normpath(src) in matches:
                raise InvalidMoveSource(src)

        shutil.move(src, dst)
        self._register_move(src, dst)

    def _register_move(self, src: str, dst: str) -> None:
        """ Registers the given (src, dst) pair as a project move. It is then
        used by the 'move_relative' method to parse and generate the moved
        paths. """

        src = os.path.relpath(src, self.location)
        dst = os.path.relpath(dst, self.location)

        with open(self.relative(cptk.constants.MOVE_FILE), 'a') as file:
            file.write(f'{src}{cptk.constants.MOVE_FILE_SEPERATOR}{dst}\n')



from __future__ import annotations

import os

import cptk.utils
from cptk import __version__
from cptk.core.collector import CommandCollector
from cptk.core.system import System


collector = CommandCollector()


collector.global_argument(
    '--version',
    action='version',
    version=f'%(prog)s {__version__}',
    help='show version message and exit',
)

collector.global_argument(
    '-W', '--work-directory',
    help='set the working directory (defaults to cwd)',
    dest='wd',
    metavar='DIRECTORY',
    default=os.getcwd(),
    type=cptk.utils.path_validator(
        dir_ok=True,
        file_ok=False,
        must_exist=True,
    ),
)

collector.global_argument(
    '-v',
    action='count',
    dest='_verbosity',
    default=0,
    help='increase the verbosity of the program',
)


collector.global_argument(
    '-y', '--yes',
    action='count',
    default=0,
    dest='_yes_count',
    help="don't ask for confirmation, assume that it is provided",
)


@collector.preprocessor
def preprocessor(namespace):
    System.set_verbosity(namespace._verbosity)
    System.set_yes(namespace._yes_count)
    return namespace


@collector.command(
    'initialize',
    aliases=['init'],
    help='initialize a new cptk project',
    description='Initialize a new cptk project in the given location, '
                'using the provided project template.',
)
@collector.argument('template', type=str)
def initialize(wd: str, template: str):
    """ Initialize a new cptk project in the given location, using the provided
    project template. """

    from cptk.local.project import LocalProject
    LocalProject.init(location=wd, template=template)


@collector.command(
    'clone',
    help='clones a problem into a local cptk project',
    description='Clones a problem from a supported website into a local cptk '
                'project. ',
)
@collector.argument(
    'url',
    help='the problem to be cloned',
    type=cptk.utils.url_validator,
)
def clone(url: str, wd: str):

    from cptk.local.project import LocalProject
    from cptk.core.system import System

    proj = LocalProject.find(wd)
    prob = proj.clone_url(url)
    proj.last = prob.location

    System.echo(prob.location)


@collector.command(
    'move',
    aliases=['mv'],
    help='moves a cloned problem to a new location',
    description='Moves a cloned problem to a new location.',
)
@collector.argument(
    'src',
    type=cptk.utils.path_validator(dir_ok=True, file_ok=False, must_exist=True),
)
@collector.argument(
    'dst',
    type=cptk.utils.path_validator(dir_ok=True, file_ok=False, must_exist=True),
)
def move(src: str, dst: str, wd: str):
    """ Moves a cloned problem to a new location. """

    from cptk.local.project import LocalProject

    proj = LocalProject.find(wd)
    proj.move(src, dst)


@collector.command('bake', aliases=['run'])
@collector.argument('name', nargs='?', default=None, type=str)
def bake(wd: str, name: str = None):

    from cptk.local.project import LocalProject
    from cptk.local.problem import LocalProblem
    from cptk.core.chef import Chef

    proj = LocalProject.find(wd)
    prob = proj.last() if name is None else LocalProblem(wd, name)
    Chef(prob).bake()


@collector.command('serve', aliases=['run'])
@collector.argument('name', nargs='?', default=None, type=str)
def serve(wd: str, name: str = None):

    from cptk.local.project import LocalProject
    from cptk.local.problem import LocalProblem
    from cptk.core.chef import Chef

    proj = LocalProject.find(wd)
    prob = proj.last() if name is None else LocalProblem(wd, name)
    Chef(prob).serve()



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from cptk.scrape import Problem
    from bs4 import BeautifulSoup


@dataclass(frozen=True)
class PageInfo:
    """ A simple dataclass that contains all information about a webpage that
    is required by the 'Website' classes to preform their actions and queries.
    """

    url: str
    data: BeautifulSoup = field(repr=False)


class Website(ABC):
    """ An abstract class that represents a website that is supported by cptk.
    Has different methods that can fetch data and information from the website.
    """

    @property
    @abstractmethod
    def name(self) -> str:
        """ The name of the website. """

    @property
    @abstractmethod
    def domain(self) -> str | list[str]:
        """ A single domain or a list of domains that are represented by this
        class. """

    @abstractmethod
    def is_problem(self, info: PageInfo) -> bool:
        """ Returns True only if the given page information describes a page
        with a single problem. """

    @abstractmethod
    def to_problem(self, info: PageInfo) -> Problem | None:
        """ Constructs and a Problem instance that represents the problem that
        is displayed in the given page. If the given page doesn't display a
        single and unique problem, returns None. """



from __future__ import annotations

from typing import TYPE_CHECKING

import pkg_resources
from bs4 import BeautifulSoup
from requests import session

from cptk.scrape import PageInfo
from cptk.scrape import Website
from cptk.utils import cptkException

if TYPE_CHECKING:
    from cptk.scrape import Problem


class InvalidClone(cptkException):
    """ Raised when the clone command is called with a 'PageInfo' instance that
    doesn't describe anything that can be cloned. """

    def __init__(self, info: PageInfo) -> None:
        self.info = info
        super().__init__(f"We don't know how to handle data from {info.url!r}")


class UnknownWebsite(cptkException):
    """ Raised when trying to fetch information from a website that is not
    registed and can't be handled by cptk. """

    def __init__(self, domain: str) -> None:
        self.domain = domain
        super().__init__(f"We don't know how to handle data from {domain!r}")


class Fetcher:

    def __init__(self) -> None:
        self.session = session()
        self._load_websites()

    def _load_websites(self) -> list[type[Website]]:
        self._websites = [
            point.load()()
            for point in pkg_resources.iter_entry_points('cptk_sites')
        ]

        self._domain_to_website = dict()
        for website in self._websites:
            domain = website.domain
            if isinstance(domain, str):
                self._domain_to_website[domain] = website
            else:
                for cur in domain:
                    self._domain_to_website[cur] = website

    def page_to_problem(self, info: PageInfo) -> Problem:
        """ Recives an arbitrary page info instance and tries to match it with
        a Website class that knows how to handle this specific website. If cptk
        doesn't find a way to parse the given webpage, it raises the
        'InvalidClone' exception. """

        for website in self._websites:
            if website.is_problem(info):
                return website.to_problem(info)
        raise InvalidClone(info)

    def to_page(self, url: str) -> PageInfo:
        """ Makes an get http/s request to the given URL and returns the result
        as a PageInfo instance. """

        if not url.startswith('http'):
            url = f'http://{url}'

        res = self.session.get(url)
        data = BeautifulSoup(res.content, 'lxml')
        return PageInfo(url, data)



from __future__ import annotations

import argparse
import inspect
from collections import defaultdict
from typing import Callable
from typing import TypeVar

F = TypeVar('F')


class CommandCollector:

    def __init__(self, *args, **kwargs) -> None:
        self._parser = argparse.ArgumentParser(*args, **kwargs)
        self._subparsers = self._parser.add_subparsers()
        self._args: dict[Callable, list[tuple]] = defaultdict(list)

        self._preprocessor = lambda v: v

        # Print the default help message if no command is provided
        self._parser.set_defaults(func=self._parser.print_help)

    def global_argument(self, *args, **kwargs) -> None:
        return self._parser.add_argument(*args, **kwargs)

    def argument(self, *args, **kwargs) -> Callable[[F], F]:
        """ Returns a decorator that adds an argument to a command function.
        A wrapper around the argparse.add_argument function. """

        # We want to support decorator stacking where the top decorator is
        # the command decorator, and the decorators after that are the argument
        # decorators, in order from top to bottom. Python executes the
        # decorators in the opposite order (from the bottom to top), and thus
        # we store all the argument decorations until we see a command decorator

        def decorator(func: F) -> F:
            self._args[func].append((args, kwargs))
            return func

        return decorator

    def command(self, *args, **kwargs) -> Callable[[F], F]:
        """ Returns a decorator that creates a new subcommand parser.
        A wrapper around the add_subparsers and add_parser functions. """

        # The command decorator should be the top decorator, which means that he
        # is the last one to be executed. All the arguments are already stored
        # in the '_args' dictionary, and that is left to do is to add them to
        # the parser in the revere order.

        def decorator(func: F) -> F:
            parser = self._subparsers.add_parser(*args, **kwargs)
            for a, k in reversed(self._args[func]):
                parser.add_argument(*a, **k)

            parser.set_defaults(func=func)
            return func

        return decorator

    def preprocessor(
        self,
        func: Callable[[argparse.Namespace], argparse.Namespace],
    ) -> Callable[[argparse.Namespace], argparse.Namespace]:
        self._preprocessor = func
        return func

    def run(self, args: list[str] = None):
        """ Executes the appropriate subcommand using the given arguments. """

        args = self._parser.parse_args(args)
        args = self._preprocessor(args)
        args: dict = vars(args)

        func: Callable = args.pop('func')
        sig = inspect.signature(func)

        return func(
            **{
                key: value for key, value in args.items()
                if key in sig.parameters.keys()
            },
        )



from __future__ import annotations

import os
from dataclasses import dataclass
from dataclasses import field
from typing import List
from typing import Optional
from typing import TYPE_CHECKING
from typing import TypeVar

import pydantic

import cptk.constants
import cptk.utils
from cptk.core.config import ConfigFileError
from cptk.core.config import ConfigFileNotFound
from cptk.core.config import Configuration
from cptk.core.system import System

if TYPE_CHECKING:
    from cptk.core.preprocessor import Preprocessor

T = TypeVar('T')


class RecipeNotFoundError(cptk.utils.cptkException):
    def __init__(self, filepath: str, msg: str) -> None:
        self.filepath = filepath
        super().__init__(msg)


class RecipeNameNotFound(RecipeNotFoundError):
    def __init__(self, filepath: str, name: str) -> None:
        self.name = name
        super().__init__(
            filepath,
            f'Recipe named {name!r} not found in recipes file {filepath!r}',
        )


class NoRecipesFound(RecipeNotFoundError):
    def __init__(self, filepath: str) -> None:
        super().__init__(
            filepath,
            f'No recipes found in recipes file {filepath!r}',
        )


class Recipe(pydantic.BaseModel):

    name: Optional[str] = None
    bake: List[str] = []
    serve: str

    @pydantic.validator('bake', pre=True)
    @classmethod
    def string_to_commands(cls, val) -> List[str]:
        if isinstance(val, str):
            return val.split('\n')
        return val

    def preprocess(self: T, processor: Preprocessor) -> type[T]:
        def parse_str(v):
            if isinstance(v, str):
                return processor.parse_string(v)
            else:
                return v

        kwargs = {
            'name': parse_str(self.name),
            'serve': parse_str(self.serve),
        }

        if self.bake:
            kwargs['bake'] = [parse_str(v) for v in self.bake]
        return type(self)(**kwargs)


class RecipesConfig(Configuration):
    recipes: List[Recipe]


@dataclass(unsafe_hash=True)
class LocalProblem:
    location: str = field(compare=True)
    name: str | None = field(compare=True, default=None)

    @classmethod
    def init(cls: type[T], location: str, recipe: Recipe) -> T:
        path = os.path.join(location, cptk.constants.RECIPE_FILE)
        config = None

        try:
            config = RecipesConfig.load(path)
        except ConfigFileNotFound:
            pass  # File doesn't exists -> we will create a new one later.
        except ConfigFileError:
            System.warn(
                f'Failed to parse existing recipe file {path!r}. '
                'Continuing the process will overwrite the file.',
            )
            ans = System.confirm('Are you sure you want to continue')
            if not ans:
                System.abort(0)

        if config is None:
            config = RecipesConfig(recipes=list())

        config.recipes.append(recipe)
        config.dump(path)

        return cls(location, recipe.name)

    @staticmethod
    def is_problem(location: str, name: str = None) -> bool:
        """ Returns True if the given directory contains a recipe with the given
        name. If a name isn't provided, returns True if a recipe file is located
        inside the given direcotry and it contains at least one recipe.  """

        path = os.path.join(location, cptk.constants.RECIPE_FILE)

        try:
            recipes = RecipesConfig.load(path)
        except ConfigFileError:
            return False

        names = {r.name for r in recipes.recipes}
        return len(names) > 0 if name is None else name in names

    @property
    def recipe(self) -> Recipe:
        path = os.path.join(self.location, cptk.constants.RECIPE_FILE)
        recipes = RecipesConfig.load(path).recipes

        if not recipes:
            raise NoRecipesFound(path)
        if self.name is None:
            return recipes[0]

        try:
            return next(r for r in recipes if r.name == self.name)
        except StopIteration:
            raise RecipeNameNotFound(path, self.name)



from __future__ import annotations

from dataclasses import dataclass
from dataclasses import field
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any
    from cptk.scrape import Website
    from datetime import datetime


@dataclass(unsafe_hash=True)
class Scraped:
    """ Any information that is scraped using cptk is represented using
    instanaces of the 'Scraped' object, or instances of classes that inherent
    from it. """

    _uid: Any = field(compare=True, hash=True)
    website: Website = field(compare=True)
    name: str = field(compare=False)
    url: str = field(compare=False)


@dataclass(unsafe_hash=True)
class Problem(Scraped):
    """ A basic problem that is scraped using cptk. Most website implement their
    own 'Problem' class which inherits from this one and can store additional
    information. """

    contest: Contest = field(compare=False)
    tests: list[Test] = field(default_factory=list, compare=False)
    time_limit: float = field(default=None, compare=False)     # in Seconds
    memory_limit: float = field(default=None, compare=False)   # in MB


@dataclass(unsafe_hash=True)
class Contest(Scraped):
    active: bool = field(default=None, compare=False)
    start_time: datetime = field(default=None, compare=False)
    end_time: datetime = field(default=None, compare=False)


@dataclass(unsafe_hash=True)
class Test:
    input: str
    expected: str = None
